from pyfreefem import FreeFemRunner

code = """
mesh Th=square(30,30);
fespace Fh(Th,P1);
Fh u,v;

DEFAULT (SOLVE_LAPLACE,0)
cout << "The value of SOLVE_LAPLACE is $SOLVE_LAPLACE." << endl;
IF SOLVE_LAPLACE
solve laplace(u,v)=
    int2d(Th)(dx(u)*dx(v)+dy(u)*dy(v))
        -int2d(Th)(v)
        +on(1,2,3,4,u=0);
plot(u,cmm="$message");
ENDIF
"""

# Small hack for running FreeFem++-nw instead of FreeFem++ command 
def cmd(self, *args, **kwargs):
    cmd = self.old_cmd(*args,**kwargs).split(" ")
    cmd[0] = "FreeFem++-nw"
    cmd = " ".join(cmd)
    return cmd

FreeFemRunner.old_cmd = FreeFemRunner.cmd
FreeFemRunner.cmd = cmd

FreeFemRunner(code,debug=1).execute({'message':'Hello world',\
                                       'SOLVE_LAPLACE':1},verbosity=0,\
                                      plot=False)



# This file is part of PyFreeFEM.
#
# PyFreeFEM is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# PyFreeFEM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# A copy of the GNU General Public License is included below.
# For further information, see <http://www.gnu.org/licenses/>.
"""PyFreeFEM input/output functions."""
    
import os
import scipy.sparse as sp
import numpy as np
import time
import sys
import struct
import subprocess
import threading
import queue
import inspect
    


try:
    import colored as col

    def colored(text, color=None, attr=None):
        if color:
            text = col.stylize(text, col.fg(color))
        if attr:
            text = col.stylize(text, col.attr(attr))
        return text
except:
    def colored(text, color=None, attr=None):
        return text


def display(message, level=0, debug=0, color=None, attr=None, end='\n',
            flag=None):
    """A function for displaying messages with a flexible tunable level of verbosity.


    :param message:  Text to be printed.
    :param level:   Level of importance of the message which will be actually 
                       printed if ``debug >= level`` or if ``flag`` is ``stdout`` or ``stderr``.
    :type level: int
    :param debug: Desired level of verbosity; the higher and the more information will  
                  be displayed.
    :type debug: int
    :param color: Color of the displayed message. Follows the convention of the     
                  `colored <https://pypi.org/project/colored/>`_ package.
    :param attr:  Formatting of the displayed message. Follows the  
                  convention of the     
                  `colored <https://pypi.org/project/colored/>`_ package.

    :param end:   character appended to the end of a message. Will remove the   
                  final line carriage return if ``end=''``.  

    :param flag:  an indicator to use to indicate if the argument ``message`` comes     
                  from ``stdout`` or ``stderr`` pipe. 
    
    .. note::   
        
       For advanced projects, it can be useful to override :func:`pyfreefem.io.display`,  
       for instance to redirect messages to a file.
    """
    if color or attr:
        message = colored(message, color, attr)
    if debug >= level or flag in ['stdout','stderr']:
        print(message, end=end, flush=True)

    
# check if stdbuf is available
def is_command_available(command):
    try:
        # Try to execute the command with `--version` to check if it is available
        result = subprocess.run([command, '--version'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        # If the command runs successfully, it's available
        return result.returncode == 0
    except FileNotFoundError:
        # If the command is not found, it raises a FileNotFoundError
        return False
WITH_STDBUF = is_command_available('stdbuf')
if not WITH_STDBUF: 
    display("[PyFreeFEM] Warning, the command stdbuf is not available. Standard output capturing may not work as expected.",color="orange_4a")

tclock = dict()
def tic(ref=0):
    global tclock
    tclock[ref] = time.time()

def toc(ref=0):
    global tclock
    return format(time.time()-tclock[ref],"0.2f")+"s"

class ExecException(Exception):
    """ An exception raised when a FreeFem++ subprocess fails"""
    def __init__(self, message, returncode, stdout, stderr, mix):
        super().__init__(message)
        self.returncode = returncode
        self.stdout = stdout
        self.stderr = stderr
        self.mix = mix

def readFFArray(ffarray : str, binary = False):
    """
    Read an FreeFEM array stored in a file ``ffarray``. 
        
    For instance, if the following ``.edp`` file is executed:

    .. code:: freefem

       real[int] table = [1, 2, 3, 4, 5];
       {
           ifstream f("file.gp");
           f << table;
       }

    Then
        
    .. code:: python

       readFFArray("file.gp") 

    returns the numpy array [1, 2, 3, 4, 5].    
        
    :param ffarray: file name of a FreeFEM array stored with ``ifstream``.
    :return: numpy array identical to ``ffarray``.
    """
    if binary:
        with open(ffarray, 'rb') as f:
            # Read the number of rows (n) and columns (m)
            n = struct.unpack('q', f.read(8))[0]  # 8 bytes for int64_t

            # Read the entire matrix data into a flat list of values
            num_elements = n 
            flat_data = struct.unpack(f'{num_elements}d', f.read(num_elements * 8))  # 'd' is for double (float64)

            # Convert the flat data to a numpy array and reshape it into an (n, m) matrix
            array = np.array(flat_data)
    else:
        with open(ffarray, "r") as f:
            return np.asarray([float(x) for line in f.readlines()[1:]
                               for x in line.split()])
    return array


def writeFFArray(A: np.ndarray, fileName: str, binary = False):
    """
    Store a numpy array into a FreeFEM array file that can be read with ``ofstream``.

    :param A:  a `numpy.array <https://numpy.org/doc/stable/reference/generated/numpy.array.html>`_ data structure
    :param fileName: an output file name to save the array ``A``.   
    """
    if binary:
        n = len(A)  
        
        with open(fileName, 'wb') as f:
            # Write the dimensions (n and m) as 64-bit integers
            f.write(struct.pack('q', n))  # Using 'q' for signed long long (64-bit)
            
            # Write the matrix data as a flat array
            f.write(struct.pack(f'={n}d', *A))  # 'd' -> float64
    else:
        text = f"{len(A)}\t\n"
        lines = [A[5*k:5*(k+1)] for k in range(0, int(np.ceil(len(A)/5)))]
        for line in lines:
            text += ''.join([f"\t  {x}" for x in line])+"\n"
        text = text[:-1]+"\t"
        with open(fileName, "w") as f:
            f.write(text)

def writeFF2DArray(A, filename):
    """
    Save a dense matrix into a FreeFEM matrix file (a real[int,int]) that can be read with ``ofstream``. 

    :param A: A `scipy.sparse.csc_matrix <https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.csc_matrix.html>`_  
              or a `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_ dense matrix.
    :param fileName: output file name for saving the matrix ``A``.  
    """ 
    # Ensure the matrix is of type float64 (similar to FreeFEM's default)
    # Get matrix dimensions
    n, m = A.shape
    
    with open(filename, 'wb') as f:
        # Write the dimensions (n and m) as 64-bit integers
        f.write(struct.pack('qq', n, m))  # Using 'q' for signed long long (64-bit)
        
        # Write the matrix data as a flat array
        f.write(struct.pack(f'={n * m}d', *A.ravel(order='F')))  # 'd' -> float64

def writeFFMatrix(A, fileName):
    """
    Save a dense or a sparse matrix into a FreeFEM matrix file that can be read with ``ofstream``. 

    :param A: A `scipy.sparse.csc_matrix <https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.csc_matrix.html>`_  
              or a `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_ dense matrix.
    :param fileName: output file name for saving the matrix ``A``.  
    """
    if isinstance(A,np.ndarray):
        preamble = " ".join(map(str,A.shape))+"\t\n"
        preamble += "\n".join(["\t   "+"   ".join(map(str,line)) for line in A.tolist()])
        preamble += "\n\t"
    else:
        (I, J, V) = sp.find(A)
        (n, m) = A.shape
        nnz = A.nnz
        preamble = f"#  HashMatrix Matrix (COO) 0x2d90200\n"
        preamble += "#    n       m        nnz     half     fortran   state  \n"
        preamble += f"{n} {m} {nnz} 0 0 0 0 \n"

        lines = [f"     {i}     {j} {d}" for (i, j, d) in zip(I, J, V)]
        preamble = preamble+"\n".join(lines)
        preamble += "\n"
    with open(fileName, "w") as f:
        f.write(preamble)
            
def readPetscMat(filepath):
    """ 
    Read a sparse matrix stored by PetsC ObjectView 
    """
    with open(filepath,"rb") as f:
        #PetscInt    MAT_FILE_CLASSID
        #PetscInt    number of rows
        #PetscInt    number of columns
        #PetscInt    total number of nonzeros
        #PetscInt    *number nonzeros in each row
        #PetscInt    *column indices of all nonzeros (starting index is zero)
        #PetscScalar *values of all nonzeros
        
        # Lire l'en-tête du fichier : MAT_FILE_CLASSID, number of rows, columns, and nonzeros
        header_format = '>4i'  # MAT_FILE_CLASSID, rows, cols, nonzeros (4 entiers de 32 bits)
        header_size = struct.calcsize(header_format)
        header_data = f.read(header_size)
        mat_file_classid, rows, cols, nnz = struct.unpack(header_format, header_data)
        print(mat_file_classid)

        # Lire le nombre de non-nuls dans chaque ligne (array de entiers 32 bits)
        # On s'attend à un tableau d'entiers de taille 'rows'
        row_nonzeros_format = f'>{rows}i'
        row_nonzeros_size = struct.calcsize(row_nonzeros_format)
        row_nonzeros_data = f.read(row_nonzeros_size)
        row_nonzeros = struct.unpack(row_nonzeros_format, row_nonzeros_data)

        # Lire les indices des colonnes des non-nuls (tableau d'entiers de 32 bits)
        col_indices_format = f'>{nnz}i'  # tableau des indices des colonnes
        col_indices_size = struct.calcsize(col_indices_format)
        col_indices_data = f.read(col_indices_size)
        col_indices = struct.unpack(col_indices_format, col_indices_data)

        # Lire les valeurs des non-nuls (tableau de doubles 64 bits)
        values_format = f'>{nnz}d'  # tableau des valeurs
        values_size = struct.calcsize(values_format)
        values_data = f.read(values_size)
        values = struct.unpack(values_format, values_data)
        
        # Calculer la position des lignes dans le tableau row_ptr
        row_ptr = np.cumsum([0] + list(row_nonzeros))  # Cumulatif des non-nuls par ligne

        # Créer la matrice CSR en utilisant les indices et les valeurs
        return sp.csr_matrix((values, col_indices, row_ptr), shape=(rows, cols))

def writePetscMatrix(matrix, file_path):
    # Récupérer les dimensions de la matrice et le nombre d'éléments non nuls
    rows, cols = matrix.shape
    nnz = matrix.nnz
    
    # Obtenir les indices de colonnes et les valeurs des éléments non nuls
    col_indices = matrix.indices
    values = matrix.data
    
    # Compter le nombre de non-nuls par ligne
    row_nonzeros = np.diff(matrix.indptr)  # La différence entre les indices de ligne

    # Ouvrir le fichier en mode binaire pour l'écriture
    with open(file_path, 'wb') as f:
        # 1. Écrire l'en-tête du fichier
        mat_file_classid = 1211216  # Exemple d'identifiant de fichier 
        header_format = '>4i'  # > pour big-endian, 4 entiers de 32 bits
        header_data = struct.pack(header_format, mat_file_classid, rows, cols, nnz)
        f.write(header_data)

        # 2. Écrire le nombre de non-nuls par ligne
        row_nonzeros_format = f'>{rows}i'
        row_nonzeros_data = struct.pack(row_nonzeros_format, *row_nonzeros)
        f.write(row_nonzeros_data)

        # 3. Écrire les indices de colonnes des non-nuls
        col_indices_format = f'>{nnz}i'
        col_indices_data = struct.pack(col_indices_format, *col_indices)
        f.write(col_indices_data)

        # 4. Écrire les valeurs des non-nuls
        values_format = f'>{nnz}d'  # Doubles 64 bits
        values_data = struct.pack(values_format, *values)
        f.write(values_data)

    print(f"Matrix saved to {file_path}")

def readFF2DArray(filepath):  
    """
    Read a matrix file generated by a FreeFEM script (a `real[int,int]`).         
    If the matrix stored by the FreeFEM script is a sparse matrix,  
    then :func:`pyfreefem.io.readFFMatrix`   
    returns a   
    `scipy.sparse.csc_matrix <https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.csc_matrix.html>`_.

    If it is a dense matrix ,   
    then :func:`pyfreefem.io.readFFMatrix` returns a `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_.
        
    :param filepath: file name of a FreeFEM matrix stored with ``ifstream``.
    :return: A `numpy.ndarray` matrix if `ffmatrix` contains a dense matrix, or a `scipy.sparse.csc_matrix` if  
             `ffmatrix` contains a sparse matrix.   
             """
    with open(filepath, 'rb') as f:
        # Read the number of rows (n) and columns (m)
        n = struct.unpack('q', f.read(8))[0]  # 8 bytes for int64_t
        m = struct.unpack('q', f.read(8))[0]  # 8 bytes for int64_t

        # Read the entire matrix data into a flat list of values
        num_elements = n * m
        #flat_data = struct.unpack(f'{num_elements}d', f.read(num_elements * 8))  # 'd' is for double (float64)
        tic()
        matrix = np.frombuffer(f.read(num_elements * 8), dtype="float64").reshape((n,m),order='F')


        # Convert the flat data to a numpy array and reshape it into an (n, m) matrix
        #tic()
        #matrix = np.asarray(flat_data).reshape((n, m),order='F')
        print("Save matrix took "+toc())
    return matrix
    
def readFFMatrixBinary(filename):
    """ 
    Matrix binary file reader from MatrixMarket plugin 
    """
    with open(filename, "rb") as f:
        # Lire l'en-tête
        header = f.readline().decode().strip()
        matcode = f.readline().decode().strip()
        M, N, nz = map(int, matcode.split())
        
        # Déterminer si la matrice est binaire
        is_binary = "%%MatrixMarketBinary" in header
        
        if is_binary:
            # Lecture des données binaires
            data = f.read(nz * (2 * 4 + 8))  # 2 entiers (indices) + 1 double (valeur) par entrée
            array = np.frombuffer(data, dtype=np.dtype([('row', 'i4'), ('col', 'i4'), ('val', 'f8')]))
            rows, cols, values = array['row'] - 1, array['col'] - 1, array['val']
        else:
            # Lecture des données en texte
            data = np.loadtxt(f, dtype=np.float64, usecols=(0, 1, 2))
            rows, cols, values = data[:, 0].astype(int) - 1, data[:, 1].astype(int) - 1, data[:, 2]
        
        # Construire la matrice creuse
        return sp.csc_matrix((values, (rows, cols)), shape=(M, N))
        
def writeFFMatrixBinary(matrix, filename):   
    matrix = matrix.tocoo()
    M, N = matrix.shape
    nz = matrix.nnz
    
    with open(filename, "wb") as f:
        # Écrire l'en-tête
        f.write(b"%%MatrixMarketBinary matrix coordinate real general\n")
        f.write(f"{M} {N} {nz}\n".encode())
        
        # Préparer les données binaires
        data = np.array(list(zip(matrix.row + 1, matrix.col + 1, matrix.data)), 
                        dtype=np.dtype([('row', 'i4'), ('col', 'i4'), ('val', 'f8')]))
        f.write(data.tobytes())
    
def readFFMatrix(ffmatrix : str):
    """
    Read a matrix file generated by a FreeFEM script. For instance,     
    if the following ``.edp`` file is run:  
        
    .. code:: freefem

       matrix table = [[1, 2, 3, 4, 5],[1, 2, 3, 4, 5]];
       {
           ifstream f("file.gp");
           f << table;
       }
    
    Then
        
    .. code:: python

       readFFMatrix("file.gp") 

    returns the numpy matrix [[1, 2, 3, 4, 5],[1, 2, 3, 4, 5]]. 
        
    If the matrix stored by the FreeFEM script is a sparse matrix, then :func:`pyfreefem.io.readFFMatrix`   
    returns a `scipy.sparse.csc_matrix <https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.csc_matrix.html>`_.

    If it is a dense matrix, then :func:`pyfreefem.io.readFFMatrix` returns a `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_.
        
    :param ffmatrix: file name of a FreeFEM matrix stored with ``ifstream``.
    :return: A `numpy.ndarray` matrix if `ffmatrix` contains a dense matrix, or a `scipy.sparse.csc_matrix` if  
             `ffmatrix` contains a sparse matrix.

    """
    with open(ffmatrix, 'r') as f:
        lines = f.readlines()
    i = []
    j = []
    data = []
    if lines[0].startswith('#  HashMatrix'):
        shape = tuple([int(x) for x in lines[2].strip().split()[:2]])
        if len(lines)>3:
            matrix = np.loadtxt(lines[3:])  
            I = matrix[:,0].astype(int) 
            J = matrix[:,1].astype(int) 
            data = matrix[:,2]
            indices = np.where(np.abs(data)>1e-15)[0]
            I, J, data = I[indices], J[indices], data[indices]
        else:
            I, J= [], []
        return sp.csc_matrix((data, (I,J)),shape=shape)
    elif len(lines[0].strip().split())==2:
        shape = tuple([int(x) for x in lines[0].strip().split()])
        data = [line.strip().split() for line in lines[1:shape[0]+1]]
        return np.asarray(data, dtype=float)
    else:
        for (k, line) in enumerate(lines[4:]):
            i.append(int(line.split()[0])-1)
            j.append(int(line.split()[1])-1)
            data.append(float(line.split()[2]))
    return sp.csc_matrix((data, (i, j)))

def enqueue_stream(stream, queue, type):
    if not WITH_STDBUF:
        for line in iter(stream.readline, ''):
            queue.put(str(type)+ line)
    else:
        for line in iter(stream.readline, b''):
            queue.put(str(type) + line.decode('utf-8', errors='replace'))
    stream.close()


def enqueue_process(process, queue):
    process.wait()
    queue.put('x')

    
def exec2(cmd, debug=0, level=1, silent=True):
    """Interface with `subprocess.Popen <https://docs.python.org/fr/3/library/subprocess.html>`_.   
       Execute a shell command ``cmd`` and pass standard output and standard errors 
       to the function :func:`pyfreefem.io.display`.
        
       :param cmd: A shell command. For instance ``cmd="FreeFem++ script.edp -v 1 -wg"``.
       :type cmd: str
        
       :param debug: An input ``debug`` parameter, tuning the desired level of verbosity.   
                     The command ``cmd`` will be displayed if ``debug>=level``
       :type debug: int

       :param level: Degree of importance of the executed ``command``.
                     The command ``cmd`` will be displayed if ``debug>=level``
       :type level: int
        
       :param silent: If set to ``True``, no standard output will be displayed.
       :type silent: boolean
        
       .. note::    
            
          By default, the standard output is displayed whatever the value of level and  
          debug because :func:`pyfreefem.io.display` prints ``stdout`` and ``stderr``   
          by watching its ``flag`` argument. Use a custom :func:`pyfreefem.io.display`    
          function to change this behaviour.

    """
    display(colored(cmd, color="indian_red_1a"), level=level, debug=debug, end='',
            flag='shell')
    # Put a line break to separate from the stdout 
    if not silent:
        display("",level=level,debug=debug,flag='')
    tic(121)
    if not WITH_STDBUF: 
        proc = subprocess.Popen(cmd,shell=True,
                                stdout = subprocess.PIPE,
                                stderr = subprocess.PIPE,
                                universal_newlines = True,
                                bufsize = 1
                                )
    else:
        proc = subprocess.Popen("stdbuf -oL "+cmd,shell=True,
                                stdout = subprocess.PIPE,
                                stderr = subprocess.PIPE
                                )
    q = queue.Queue()
    to = threading.Thread(target=enqueue_stream, args=(proc.stdout, q,1))
    te = threading.Thread(target=enqueue_stream, args=(proc.stderr, q,2))
    tp = threading.Thread(target=enqueue_process, args=(proc, q))
    te.start()
    to.start()
    tp.start()

    stdout = "" 
    stderr = "" 
    mix = "" 
    while True:
        line = q.get()
        if line[0] == 'x':
            break
        if line[0] == '1':
            line = line[1:]
            stdout += line
            mix += line 
            if not silent:
                display(line,level=level+1,debug=debug,end='',flag='stdout')
        if line[0] == '2':
            line = line[1:]
            stderr += line
            mix += line 
            if not silent:
                display(line,level=level+1,debug=debug,attr="dim",end='',flag='stderr')
    tp.join()
    te.join()
    to.join()
    if mix and not silent:
        if not mix.endswith('\n'):
            display("",level,debug)
        display("Finished in",level,debug,color="cyan",end="",flag="")
    display(' ('+toc(121)+')', level=level, debug=debug,color="cyan",
            flag="time")

    if proc.returncode != 0:
        raise ExecException('Error : the process "'
                        + colored(cmd, "red")
                        + '" failed with return code '+str(proc.returncode)
                        + ".",proc.returncode,stdout,stderr,mix)
    return proc.returncode, stdout, stderr, mix
    
def make_dictionary(*args):
    frame = inspect.currentframe().f_back  # Get the caller's frame
    names = {id(v): k for k, v in frame.f_locals.items()}  # Map object ids to variable names
    return {names.get(id(arg), f'unknown_{i}'): arg for i, arg in enumerate(args)}

def extract(d, *keys):  
    if len(keys)==1:    
        return d[keys[0]]
    else:
        return tuple(d[k] for k in keys)

def get_root(): 
    file = inspect.stack()[1][1]
    return os.path.dirname(file)

def tic(ref=0):
    global tclock
    tclock[ref] = time.time()


def toc(ref=0):
    global tclock
    return format(time.time()-tclock[ref], "0.2f")+"s"



# This file is part of PyFreeFEM.
#
# PyFreeFEM is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# PyFreeFEM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# A copy of the GNU General Public License is included below.
# For further information, see <http://www.gnu.org/licenses/>.
from pyfreefem.io import readPetscMat, writePetscMatrix
from pyfreefem import FreeFemRunner
import numpy as np
import scipy as sp
import scipy.sparse as spa

code = """
IMPORT "io.edp" 

mesh Th=square(3,3);
fespace Fh(Th,P1);
varf laplace(u,v)=int2d(Th)(2*u*v+dx(u)*dx(v)+dy(u)*dy(v));
matrix A = laplace(Fh,Fh,tgv=-2);
    
exportMatrix(A); 
    
Mat APetsc(A); 
ObjectView(APetsc, format = "binary", name = "matpetsc_A");
"""
    
exports = FreeFemRunner(code,debug=3).execute(with_mpi=1,verbosity=0)

APetsc = readPetscMat("matpetsc_A")
A = exports['A']
print(APetsc)
print(APetsc-A)
    
writePetscMatrix(A, "matpetsc_B")
code = """
IMPORT "io.edp" 

Mat APetsc; 
MatLoad(APetsc, format = "binary", name = "matpetsc_B");

ObjectView(APetsc,format="info"); 
ObjectView(APetsc, format = "binary", name = "matpetsc_C");
"""
FreeFemRunner(code,debug=3).execute(with_mpi=True,verbosity=0); 
CPetsc = readPetscMat("matpetsc_C")
print(APetsc-CPetsc)



# This file is part of PyFreeFEM.
#
# PyFreeFEM is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# PyFreeFEM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# A copy of the GNU General Public License is included below.
# For further information, see <http://www.gnu.org/licenses/>.
from pyfreefem import FreeFemRunner, extract
try:
    import pymedit  
    WITH_PYMEDIT = True
except:
    WITH_PYMEDIT = False

code = """
IMPORT "io.edp"

real var = pi;
real[int] arr = [1,2,3,4];

mesh Th = square(10,10);
fespace Fh1(Th,P1);

macro grad(u) [dx(u),dy(u)]//
varf laplace(u,v) = int2d(Th)(u*v+grad(u)'*grad(v));
varf rhs(u,v) = int1d(Th,1)(v);

matrix A = laplace(Fh1,Fh1);
real[int] f= rhs(0,Fh1);

real[int] u=A^-1*f;
    
real[int,int] B = [[1,2,3],[4,5,6]]; 


dispVar(var);
exportVar(var);
export2DArray(B); 

IF PYMEDIT
exportMesh(Th);
ENDIF   

exportArray(arr);
exportArray(f);
exportArray(u);
exportMatrix(A);
"""

config = dict(PYMEDIT=int(WITH_PYMEDIT))
exports = FreeFemRunner(code).execute(config=config,verbosity=0)

if __name__=="__main__":
    print('var='+str(exports['var']))
    print('arr='+str(exports['arr']))
    print('A=',exports['A'])
    print('B=',exports['B'])
    print('f=',exports['f'])
    if WITH_PYMEDIT:
        import matplotlib.pyplot as plt
        plt.ion()
        exports['Th'].plot()

        from pymedit import P1Function
        Th = exports['Th']
        u = P1Function(Th, exports['u'])
        u.plot(title="u")
        input("Press any key")







# Copyright 2018-2020 CNRS, Ecole Polytechnique and Safran.
#
# This file is part of PyFreeFEM.
#
# PyFreeFEM is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# PyFreeFEM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# A copy of the GNU General Public License is included below.
# For further information, see <http://www.gnu.org/licenses/>.
import sys
if len(sys.argv)==3 and sys.argv[1]=='--debug':
    debug = int(sys.argv[2])
else:
    debug = 0
from pyfreefem import FreeFemRunner
from pyfreefem.io import colored

code = """
mesh Th=square(30,30);
fespace Fh(Th,P1);
Fh u,v;

DEFAULT (SOLVE_LAPLACE,0)
cout << "The value of SOLVE_LAPLACE is $SOLVE_LAPLACE." << endl;
IF SOLVE_LAPLACE
solve laplace(u,v)=
    int2d(Th)(dx(u)*dx(v)+dy(u)*dy(v))
        -int2d(Th)(v)
        +on(1,2,3,4,u=0);
plot(u,cmm="$message");
ENDIF
"""

runner = FreeFemRunner(code,debug=debug)    
    
if __name__=='__main__':
    print(colored("Running with verbosity=1:",attr="bold"))
    runner.execute({'message':'Hello world',\
                    'SOLVE_LAPLACE':1},verbosity=1,\
                   plot=True)

    print(colored("\nRunning with verbosity=0:",attr="bold"))
    runner.execute({'message':'Hello world',\
                    'SOLVE_LAPLACE':1},verbosity=0,\
                   plot=False)

    print(colored("\nRunning with default verbosity (-1):",attr="bold"))
    runner.execute({'message':'Hello world',\
                    'SOLVE_LAPLACE':1},\
                   plot=False)



# PyFreeFEM is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# PyFreeFEM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# A copy of the GNU General Public License is included below.
# For further information, see <http://www.gnu.org/licenses/>.
from pyfreefem.io import readFFArray, writeFFArray
from pyfreefem import FreeFemRunner
import numpy as np

code = """
real[int] table = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12];
{
    ofstream f("$RUNDIR/ex7_x.gp");
    f << table;
}
cout << "FreeFEM : saved ex7_x.gp" << endl;
cout << table << endl;
"""
with FreeFemRunner(code) as runner: 
    runner.execute(verbosity=0)
    x = readFFArray(runner.run_dir+"/ex7_x.gp")

print(f"Python : read ex7_x.gp \nnumpy array x={x}")
x2 = np.flip(x,0)

code = """
int n;
{
    ifstream f("$RUNDIR/ex7_x2.gp");
    f >> n;
}
real[int] table(n);
{ 
    ifstream f("$RUNDIR/ex7_x2.gp");
    f >> table;
}
cout << "FreeFEM : read ex7_x2.gp " << endl << table << endl;
"""
with FreeFemRunner(code) as runner: 
    writeFFArray(x2, runner.run_dir+"/ex7_x2.gp")
    print(f"Python : saved ex7_x2.gp \nnumpy array x2={x2}")
    runner.execute(verbosity=0)
    rets = runner.rets



# This file is part of PyFreeFEM.
#
# PyFreeFEM is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# PyFreeFEM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# A copy of the GNU General Public License is included below.
# For further information, see <http://www.gnu.org/licenses/>.
from pyfreefem import FreeFemRunner

import numpy as np
from pymedit import square
import scipy.sparse as sp
    
x = np.pi   
arr = np.asarray([1,2,3])
B = np.asarray([[1,2,3],[4,5,6]])
Th = square(10,10)
A = sp.diags([1,2,3,4])

code = """
IMPORT "io.edp"

mesh Th=importMesh("Th");
plot(Th,cmm="Th");

real x = importVar("x");
dispVar(x);

real[int] arr = importArray("arr"); 
dispArray(arr);

matrix A = importMatrix("A");
dispMatrix(A);
    
real[int,int] B = import2DArray("B"); 
disp2DArray(B); 
"""

runner = FreeFemRunner(code) 
    
runner.import_variables(x, arr, Th, A, B)
runner.execute(verbosity=0, plot=True)
    
result = runner.rets[1]







# This file is part of PyFreeFEM.
#
# PyFreeFEM is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# PyFreeFEM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# A copy of the GNU General Public License is included below.
# For further information, see <http://www.gnu.org/licenses/>.
from pyfreefem.io import readFFMatrixBinary, writeFFMatrixBinary
from pyfreefem import FreeFemRunner
import numpy as np
import scipy as sp
import scipy.sparse as spa

code = """
IMPORT "io.edp" 

mesh Th=square(3,3);
fespace Fh(Th,P1);
varf laplace(u,v)=int2d(Th)(2*u*v+dx(u)*dx(v)+dy(u)*dy(v));
matrix A = laplace(Fh,Fh,tgv=-2);
matrix B = [[1,2,3],[4,5,6]]; 
    
exportMatrix(A); 
exportMatrix(B); 
    
savemtx("A.mtx",A,true); 
savemtx("B.mtx",B,true); 
"""
    
exports = FreeFemRunner(code,debug=3).execute(with_mpi=1,verbosity=0)

import numpy as np
import struct
from scipy.sparse import coo_matrix

A = readFFMatrixBinary("A.mtx")
B = readFFMatrixBinary("B.mtx")
writeFFMatrixBinary(A,"A.mtx")
writeFFMatrixBinary(B,"B.mtx")

code = """
IMPORT "io.edp" 

matrix A, B; 
readmtx("A.mtx",A,true); 
readmtx("B.mtx",B,true); 
    
exportMatrix(A); 
exportMatrix(B); 
"""

exports = FreeFemRunner(code,debug=3).execute(with_mpi=1,verbosity=0)
print("A-exports['A']=",A-exports['A'])
print("exports['B']=",exports['B'].todense())
    

# Exemple d'utilisation
# matrix = read_mtx_binary("matrix.mtx")
# print(matrix)



# This file is part of PyFreeFEM.
#
# PyFreeFEM is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# PyFreeFEM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# A copy of the GNU General Public License is included below.
# For further information, see <http://www.gnu.org/licenses/>.
    

"""Module FreeFemRunner"""

from os import path
from .preprocessor import Preprocessor
import os.path
import shutil
from .io import display, exec2, ExecException, readFFMatrix, readFFArray, writeFF2DArray,   \
                   writeFFArray, writeFFMatrix, make_dictionary, readFF2DArray, writeFFMatrixBinary, \
                   readFFMatrixBinary, tic, toc
import tempfile
import numpy as np
import scipy.sparse as sp
import pymedit
import inspect
    

class FreeFemRunner:
    """A class to run a FreeFem ``.edp`` scripts    
    supporting PyFreeFEM meta-language.   

    **Usage**

    * Load a single file .edp file
        
      .. code:: python    

         runner=FreeFemRunner('solveState.edp')
    

    * Load a list of files (to be assembled consecutively)


      .. code:: python    

         runner=FreeFemRunner(['params.edp','solveState.edp'])  


    * Load a single file with an updated configuration


      .. code:: python    

         runner=FreeFemRunner('solveState.edp',{'ITER':'0010'}) 


    * Load raw code


      .. code:: python    

         code = "mesh Th=square(10,10);"
         runner=FreeFemRunner(code) 


    * Load a pyfreefem.Preprocessor object


      .. code:: python    

         preproc = Preprocessor('solveState.edp')
         runner = FreeFemRunner(preproc)


    Then execute the code with :meth:`pyfreefem.FreeFemRunner.execute`:

    .. code:: python    

       runner.execute();
       runner.execute({'Re':30}); # Update magic variable value


    :parameter script:    Either:   

                          * raw FreeFEM code   (e.g. ``script="mesh Th=square(3,3)"``) 

                          * name of ``.edp`` script file (e.g. ``script="script.edp"``) 

                          * list of names of ``.edp`` script files to be appended   
                            in the generated executable ``.edp`` script     
                            (e.g. ``script=['script1.edp','script2.edp','script3.edp']``)

                          * instance of :mod:`pyfreefem.preprocessor.Preprocessor`

    :parameter config:  a default assignment of magic variables
                        which can be modified later on during the execute() 
                        operation.  
                            
                        Example: ```{'N':10, 'RES':'high'}```.  
                            
                        .. note::   
                            
                           The values of the magic variables are converted to   
                           strings in the interpreted executable ``.edp`` script.   

    :type config:      dict

    :parameter run_dir:    running directory where the executable script is written    
                          when calling :meth:`pyfreefem.FreeFemRunner.execute`.   
                          If ``run_dir`` is ``None``, then a temporary directory     
                          is created for this purpose and deleted after execution.     
                          The temporary name directory can be retrieved by using the   
                          ``run_dir`` attribute of the :class:`pyfreefem.FreeFemRunner`    
                          instance or from the ``$RUNDIR`` magic variable.

    :parameter run_file:   Name of the generated executable ``.edp`` script. If   
                          ``run_file`` is ``None``, then ``run_file`` is set to: 
                             
                          * ``run_file=script`` if ``script`` is the name of a file 

                          * ``run_file=script[0]`` if ``script`` is a list and   
                            ``script[0]`` is the name of a file 
                             
                          * ``run.edp`` if none of the previous case applies.

    :parameter debug:       More and more debugging information with a flexible level of verbosity    
                        are displayed as this parameter is set to a larger and larger value.
                        Details of the parsing operation are displayed if 
                        ``debug>=10`` (default ``debug=0``).
    :type debug:        int

    :parameter plot: (default: ``False``). enables  
                      ``FreeFem++`` graphics (``-wg`` option) which are disabled 
                      by default.

    :parameter macro_files: List of enhanced .edp macro files
                            supporting meta-instructions
                            which are also parsed and placed in the 
                            run folder of the final executable. For instance:   
                                
                            .. code:: python    
                                
                               FreeFemRunner('script.edp',macro_files=['macros.edp','params.edp'])

    """
    
    def __init__(self, script, config=dict(), run_dir=None,
                 run_file=None, debug=0,  plot=False,
                 macro_files=None):
        """
        """
        self.freefemFiles = dict()
        self.debug = int(debug)
        self.run_file = run_file
        self.run_time = -1
        self.exports = []
        self.plot = plot

        #: Captured output of FreeFem++ process.   
        #: After calling :meth:`pyfreefem.FreeFemRunner.execute`,
        #: `rets` is a tuple    
        #: containing:  
        #:  
        #: * rets[0]: the return code   
        #: * rets[1]: stdout    
        #: * rets[2]: stderr    
        #: * rets[3]: the whole standard output (mix of stdout and stderr)
        self.rets = tuple() 

        # create temporary directory if needed
        self.__context__ = False

        if run_dir is None:
            self.tempdir = tempfile.TemporaryDirectory(prefix="pyfreefem_")
            self.run_dir = self.tempdir.name
        else:   
            self.run_dir = str(run_dir)

        if not os.path.exists(self.run_dir):
            os.makedirs(self.run_dir)
            display("Create "+self.run_dir, level=10, debug=self.debug,
                    color="magenta")
        self.ffexport_dir = self.run_dir+'/ffexport'
        self.ffimport_dir = self.run_dir+'/ffimport'

        if os.name == 'nt':
            # If windows, do not use backslashes for directory
            self.run_dir = self.run_dir.replace('\\','/')
            self.ffexport_dir = self.ffexport_dir.replace('\\','/')
            self.ffimport_dir = self.ffimport_dir.replace('\\','/')

        if isinstance(script, Preprocessor):
            self.preprocessor = script
        else:
            self.preprocessor = Preprocessor(script, config, debug=self.debug-10)
                
        #: Dictionary of magic variables.   
        #: This dictionary is updated after calling :meth:`pyfreefem.FreeFemRunner.execute`.
        self.config = config


        if self.run_file is None:
            if isinstance(script,list) and os.path.isfile(script[0]):
                self.run_file = os.path.basename(script[0])
            elif os.path.isfile(script):
                self.run_file = os.path.basename(script)
            else:
                self.run_file = "run.edp"

        if macro_files is None:     
            macro_files = []
        self.macro_files = macro_files
        self.macro_files.append(os.path.dirname(__file__)+"/edp/io.edp")
        
    def import_variables(self, *args, **kwargs):
        """Import variables making them accessible from FreeFEM script  
            
        :parameter args: variables directly entered. They will be accessible in FreeFEM     
                         through the same name.     
                            
                         .. code:: python   
                                
                            runner.import_variables(x,M);   
                                
                         .. code:: freefem  
                            
                            IMPORT "io.edp"     
                            real x=importVar("x");    
                            matrix M=importMatrix("M");     

        :parameter kwargs: provide the variables in the form of a dictionary.   
                           In that case, the variable bear the names corresponding to the   
                           keys of the dictionary. 
                            
                         .. code:: python   
                                
                            runner.import_variables(x=x1,mat=M);   
                                
                         .. code:: freefem  
                            
                            IMPORT "io.edp"     
                            real x=importVar("x");    // The content of x1
                            matrix M=importMatrix("mat"); // The content of M
        """
        frame = inspect.currentframe().f_back  # Get the caller's frame
        names = {id(v): k for k, v in frame.f_locals.items()}  # Map object ids to variable names
        extra_vars = {names.get(id(arg), f'unknown_{i}'): arg for i, arg in enumerate(args)}
        kwargs.update(extra_vars)
        # Clean the ffimport dir 
        #if os.path.exists(self.ffimport_dir):
        #    shutil.rmtree(self.ffimport_dir, ignore_errors=True)
        os.makedirs(self.ffimport_dir, exist_ok=True)
        display("Reset directory "+self.ffimport_dir, level=10, debug=self.debug, 
                color="magenta")
        tic()
        # Import python variables to FreeFEM
        for varname, var in kwargs.items():
            if isinstance(var,float) or isinstance(var,int):   
                with open(self.ffimport_dir+"/var_"+varname,"w") as f:  
                    f.write(str(var)+"\n")
            elif isinstance(var,np.ndarray) and len(var.shape)==1:        
                writeFFArray(var, self.ffimport_dir+"/array_"+varname, binary=True)
            elif isinstance(var,np.ndarray):    
                writeFF2DArray(var, self.ffimport_dir+"/2darray_"+varname)
            elif sp.issparse(var):    
                writeFFMatrixBinary(var, self.ffimport_dir+"/matrix_"+varname)
            elif isinstance(var,pymedit.mesh.Mesh):
                var.save(self.ffimport_dir+"/mesh_"+varname+".mesh")
            elif isinstance(var,pymedit.mesh3D.Mesh3D):
                var.save(self.ffimport_dir+"/mesh3D_"+varname+".meshb")
            else:   
                raise Exception("Error, type "+str(type(var)) + " is unknown for pyfreefem import." 
                                " Supported types are float, np.ndarray, pymedit.mesh.Mesh and "    
                                "pymedit.mesh3D.Mesh3D.")
        display("Imported files in "+toc(),level=3,debug=self.debug, color="orange_4a")

    def parse(self, config):    
        """Returns the parsed executable ``.edp`` script without executing it.

           :parameter config:  a default assignment of magic variables
                               which can be modified later on during the execute() 
                               operation.  
                                   
                               Example: ```{'N':10, 'RES':'high'}```.  
                                   
                               .. note::   
                                   
                                  The values of the magic variables are converted to   
                                  strings in the interpreted executable ``.edp`` script.   
            :return: a parsed ``.edp`` script.  For instance:   
                
                     .. code:: python   
                        
                        script="mesh Th=square($N,$N);" 
                        FreeFemRunner(script).parse({'N':10})
                            
                     .. code:: freefem 
                        
                        mesh Th=square(10,10);
                
          
        """
        return self.preprocessor.parse(config)

    def _write_edp_file(self, config=dict()):
        """
        Write the parsed .edp file in the running directory.

        Arguments
        ---------

            config      : a dictionary of magic variable which updates the 
                          default FreeFemRunner.config assignment.
        """
        target_file = path.join(self.run_dir, self.run_file)

        code = self.parse(config)
        
        f = open(target_file, "w")
        f.write(code)
        f.close()
        display("Write "+target_file, level=10, debug=self.debug,
                color="magenta")

        if self.preprocessor.imports:
            self.macro_files += self.preprocessor.imports

        if self.macro_files:
            for mf in self.macro_files:
                file_name = os.path.split(mf)[1]
                with open(self.run_dir+"/"+file_name, "w") as f:
                    f.write(Preprocessor(mf, self.config, debug=self.debug-10).parse(config))
        return target_file

    def __enter__(self):
        if not os.path.exists(self.run_dir):
            os.makedirs(self.run_dir)
            display("Create "+self.run_dir, level=10, debug=self.debug,
                    color="magenta")
        return self

    def __exit__(self, type, value, traceback):
        pass
            

    def execute(self, config=dict(), debug = None, verbosity = -1,  
                ncpu = 1, with_mpi = False, plot = None, level = 1, 
                extra_args="", import_variables=None):
        """
        Parse the script with the input config, save the code in an ``.edp`` file
        and execute FreeFEM on it.

        .. code:: python

            runner = FreeFemRunner('solveState.edp')
            runner.execute();
            runner.execute({"ITER" : "0024"});

        :parameter config: Dictionary of magic variable values. This    
                           parameter takes precedence over the default configuration    
                           specified in the :class:`pyfreefem.FreeFemRunner` constructor.
                            
                           .. warning::

                              A "SET" instruction in the .edp file
                              always has the precedence over the values specified
                              by `config`. See :ref:`set`.

        :type config: dict

        :parameter debug:  debugging parameter which takes precedence over the default value    
                           specified in the  :class:`pyfreefem.FreeFemRunner` constructor.  
        :type debug:   int

        :parameter verbosity: (default ``-1``). If ``verbosity`` is set to a value    
                              ``0`` or greater, the standard output of the ``FreeFem++``    
                              command will be displayed during execution with the   
                              level of verbosity corresponding to the option ``-v verbosity``.  
                              No display is shown if ``verbosity<0``.   

        :type verbosity: int

        :parameter ncpu: number of compute nodes to be used (option ``-np`` of ``ff-mpirun``). 
        :type ncpu: int

        :parameter with_mpi: execute the interpreted ``.edp`` script with ``ff-mpirun``.

                              .. code:: python    

                                 # Run on 4 cpus with ff-mpirun -np 4
                                 runner.execute(ncpu=4); 
                                 # Run on 1 cpu with ff-mpirun -np 1
                                 runner.execute(ncpu=1,with_mpi=1);

                              Note that in that case, a magic variable ``WITH_MPI`` is automatically
                              assigned and set to 1, which is useful to make adaptations in the 
                              source code, e.g. 
                                    
                              .. code:: freefem

                                 IF WITH_MPI
                                 /* Instructions to do if the code is parallel */
                                 ELSE
                                 /* Instructions to do if the code is not parallel */
                                 int mpirank = 0; 
                                 int mpisize = 1;
                                 ENDIF

        :type with_mpi: boolean
            
        :parameter plot: Enables or disables  ``FreeFem++`` graphics (``-wg`` option) which are disabled 
                         by default.    This parameter takes precedence over    
                         the ``plot`` parameter of the  
                         :class:`pyfreefem.FreeFemRunner` constructor.

        :type plot: boolean

        :parameter level: set a custom level for PyFreeFEM debugging informations.  
                          Namely, the executed ``FreeFem++`` command and further    
                          PyFreeFEM operations will be shown if ``debug>=level``.
                    
        :parameter extra_args: (default: empty). Extra arguments to pass to FreeFEM command (for instance, ``-log_view``)

        :return exports:    a dictionary containing FreeFEM exported variables. 
                            See :ref:`export`.
        :rtype: dict: 
        """
        if import_variables:    
            self.import_variables(**import_variables)
        selfdebug = self.debug 
        debug = self.debug if debug is None else debug  
        self.debug = debug
        plot = self.plot if plot is None else plot

        config.update({'RUNDIR':self.run_dir})
        config.update({'FFEXPORTDIR':self.ffexport_dir})
        config.update({'FFIMPORTDIR':self.ffimport_dir})

        config.update({'DEBUG':debug})
        if ncpu > 1 or with_mpi:
            config.update({'WITH_MPI': 1})
        else:
            config.update({'WITH_MPI': 0})
        self.config.update(config)
        target_file = self._write_edp_file(config)
        silent = verbosity < 0
        if silent:
            verbosity = -1

        self.exports = dict()
        if not os.path.exists(self.run_dir):
            os.makedirs(self.run_dir)

        # Clean the ffexport_dir 
        if os.path.exists(self.ffexport_dir):
            shutil.rmtree(self.ffexport_dir, ignore_errors=True)
        os.makedirs(self.ffexport_dir)
        display("Reset directory "+self.ffexport_dir, level=10, debug=self.debug, 
                color="magenta")


        try:
            returncode, stdout, stderr, mix = \
                exec2(self.cmd(target_file, ncpu=ncpu, verbosity= verbosity, with_mpi=with_mpi,
                               plot=plot,extra_args=extra_args),
                       debug=debug, level=level, silent=silent)
        except ExecException as e:
            display(e.args[0], level = 0, debug = 0)
            display('\n'.join(e.mix.splitlines()[-40:]), level=0, debug=0)
            e.args = []
            raise e

        exportfiles = os.listdir(self.ffexport_dir)
        tic()
        for file in exportfiles:
            if file.startswith('mesh_') and not file.endswith('.gmsh'):
                from pymedit import Mesh
                self.exports[file[5:-5]] = Mesh(self.ffexport_dir+"/"+file) 
            if file.startswith('mesh3D_') and not file.endswith('.gmsh'):
                from pymedit import Mesh3D
                self.exports[file[7:-5]] = Mesh3D(self.ffexport_dir+"/"+file) 
            if file.startswith('var_'): 
                with open(self.ffexport_dir+'/'+file) as f:
                    self.exports[file[4:]] = float(f.read())
            if file.startswith('array_'):
                self.exports[file[6:]] = readFFArray(self.ffexport_dir+'/'+file, binary=True)
            if file.startswith('arrayInt_'):
                self.exports[file[9:]] = readFFArray(self.ffexport_dir+'/'+file, binary=True).astype(int)
            if file.startswith('matrix_'):
                self.exports[file[7:]] = readFFMatrixBinary(self.ffexport_dir+'/'+file)
            if file.startswith('2darray_'):
                self.exports[file[8:]] = readFF2DArray(self.ffexport_dir+'/'+file)
        display("Exported files in "+toc(),level=3,debug=self.debug,color="orange_4a")

        self.debug = selfdebug 

        self.rets =(returncode, stdout, stderr, mix)
        return self.exports


    def cmd(self, target_file, ncpu=1, with_mpi=False, verbosity = None, plot=False, extra_args=""):
        """Returns the shell command that is to be run when calling the method :meth:`pyfreefem.FreeFemRunner.execute`.  
            
        :parameter target_file: name of the final executable file.  

        :parameter with_mpi: (default: ``False``). If set to ``True``, then     
                             :meth:`pyfreefem.FreeFemRunner.execute` will call ``ff-mpirun``.

        :type with_mpi: boolean

        :parameter ncpu: number of compute nodes to be used (option ``-np`` of ``ff-mpirun``). 

        :parameter verbosity: option '-v` of FreeFEM. This option is not specified if   
                              ``verbosity`` is ``None``.

        :parameter plot: (default: ``False``). If set to ``True``, then     
                         :meth:`pyfreefem.FreeFemRunner.execute` will use the   
                         ``-wg`` option (with graphics). Otherwise,     
                         FreeFem will be called with ``-nw`` option.

        :parameter extra_args: (default: empty). Extra arguments to pass to FreeFEM command (for instance, ``-log_view``)

        :type plot: boolean
            
        :returns: Shell command for executing FreeFEM with ``target_file`` input 
                  script.
        """
        if ncpu > 1 or with_mpi:    
            cmd = f"ff-mpirun -np {ncpu}"
        else:
            cmd = "FreeFem++"
        cmd += " " + target_file
        if not verbosity is None:
            cmd = cmd+" -v "+str(int(verbosity))
        if plot:    
            cmd = cmd+" -wg"
        elif not with_mpi:
            cmd += " -nw"
        if extra_args:  
            cmd += " "+extra_args
        return cmd



# This file is part of PyFreeFEM.
#
# PyFreeFEM is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# PyFreeFEM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# A copy of the GNU General Public License is included below.
# For further information, see <http://www.gnu.org/licenses/>.
import os
from pyfreefem import Preprocessor, FreeFemRunner
from pyfreefem.io import readFFArray, writeFFArray, readFFMatrix, writeFFMatrix
import numpy as np
import scipy as sp
try:
    import pymedit  
    WITH_PYMEDIT = True
except:
    WITH_PYMEDIT = False

examples_folder = os.path.split(os.path.split(__file__)[0])[0]+'/pyfreefem/examples'

def test_ex00_substitution():    
    from pyfreefem.examples.ex00_substitution import parsed
    expected ='real Re=100;\nreal bigPe=30;//Accolades needed'
    assert parsed == expected

def test_ex01_default_set():
    from pyfreefem.examples.ex01_default_set import parsed_default, config_default, parsed_variant, config_variant
    assert parsed_default == 'real Re=10;\nreal Pe=30;\nreal ratio=0.3333333333333333;\n'
    assert parsed_variant == 'real Re=100;\nreal Pe=30;\nreal ratio=3.3333333333333335;\n'
    assert config_default == {'Re': 10, 'Pe': 30, 'ratioRePe': 0.3333333333333333}  
    assert config_variant == {'Pe': 30, 'Re': 100, 'ratioRePe': 3.3333333333333335}
        
def test_ex02_set_textvar(): 
    from pyfreefem.examples.ex02_set_textvar import preproc 
    assert preproc.config['params.mmg2d']=='Parameters\n1\n \n10 Edges 0.0001 0.001 1e-05\n'
    
def test_ex03_ifeq():    
    from pyfreefem.examples.ex03_ifeq import parsed_default, parsed_P2, parsed_P1b
    expected_default = '\ncout << "Solving with P1 finite elements." << endl;\n\nmesh Th=square(30,30);\nfespace Fh(Th,P1);\nFh u,v;\nsolve laplace(u,v)=\n    int2d(Th)(dx(u)*dx(v)+dy(u)*dy(v))\n        -int2d(Th)(v)\n        +on(1,2,3,4,u=0);\n\ncout << "ORDER_OK has been set." << endl;\n'
    expected_P2 = '\n\ncout << "Solving with P2 finite elements." << endl;\n\nmesh Th=square(30,30);\nfespace Fh(Th,P2);\nFh u,v;\nsolve laplace(u,v)=\n    int2d(Th)(dx(u)*dx(v)+dy(u)*dy(v))\n        -int2d(Th)(v)\n        +on(1,2,3,4,u=0);\n\ncout << "ORDER_OK has been set." << endl;\n' 
    expected_P1b = '\n\ncout << "The specified finite element order should be P1 or P2." << endl;\n\n'  
    assert parsed_default == expected_default  
    assert parsed_P2 == expected_P2
    assert parsed_P1b == expected_P1b

def test_ex04_for(): 
    from pyfreefem.examples.ex04_for import parsed
    assert parsed == '//Create a square mesh of size 1x1\nmesh Th1 = square(10,10,[1*x,1*y]);\n\n//Create a square mesh of size 2x2\nmesh Th2 = square(10,10,[2*x,2*y]);\n\n//Create a square mesh of size 3x3\nmesh Th3 = square(10,10,[3*x,3*y]);\n\n//Create a square mesh of size 4x4\nmesh Th4 = square(10,10,[4*x,4*y]);\n\n//Create a square mesh of size 5x5\nmesh Th5 = square(10,10,[5*x,5*y]);\n\n//Create a square mesh of size 6x6\nmesh Th6 = square(10,10,[6*x,6*y]);\n\n//Create a square mesh of size 7x7\nmesh Th7 = square(10,10,[7*x,7*y]);\n\n//Create a square mesh of size 8x8\nmesh Th8 = square(10,10,[8*x,8*y]);\n\n//Create a square mesh of size 9x9\nmesh Th9 = square(10,10,[9*x,9*y]);\n\n'
        
def test_ex05_magic_comment():   
    from pyfreefem.examples.ex05_magic_comment import parsed
    assert parsed == '\n\n// Display the Reynolds number\ncout << "Re=100;" << endl;\n'
        
def test_ex06_double_backslash():    
    from pyfreefem.examples.ex06_double_backslash import parsed
    assert parsed == 'mesh Th=square(10,10);\nfespace Fh(Th,P1);\nFh u,v;\n\nsolve laplace(u,v)=\n    int2d(Th)(dx(u)*dx(v)+dy(u)*dy(v))\n        -int2d(Th)(v); \n        //This final semicolumn character will be put \n        //at the end of the last previous line\n\n'

def test_ex07_read_write_FFArray():  
    from pyfreefem.examples.ex07_read_write_FFArray import x, rets
    assert np.allclose(list(range(1,13)),x,atol=0)
    assert rets[1] == 'FreeFEM : read ex7_x2.gp \n12\t\n\t 12\t 11\t 10\t  9\t  8\n\t  7\t  6\t  5\t  4\t  3\n\t  2\t  1\t\n'
            
def test_ex08_read_write_FFMatrix(): 
    from pyfreefem.examples.ex08_read_write_FFMatrix import A, rets
    expected = np.array([[ 4/3, -5/12, -5/12,  1/6],
        [-5/12,  7/6,  0.        , -5/12],
        [-5/12,  0.        ,  7/6, -5/12],
        [ 1/6, -5/12, -5/12,  4/3]])
    assert np.allclose(A.todense(),expected,atol=0)

def test_ex09_freefemrunner_env():  
    from pyfreefem.examples.ex09_freefemrunner_env import result
    assert np.allclose(result,0.6666666666666695,1e-10)


def test_ex10_io_export():
    from pyfreefem.examples.ex10_io_export import exports
    assert np.allclose(exports['A'].trace(),400.5,atol=1e-10)
    assert np.allclose(exports['arr'],np.array([1,2,3,4]),atol=0)
    assert np.allclose(sum(exports['u']),121.89328045616362,atol=1e-10)
    assert np.allclose(sum(exports['f']*exports['u']),1.3121929992942643,atol=1e-10)
    assert np.allclose(exports['var'],np.pi,1e-10)
    if WITH_PYMEDIT:    
        assert exports['Th'].nv == 121
            
def test_ex11_io_import():
    from pyfreefem.examples.ex11_io_import import result
    res = result
    assert result[-140:]==' 0 0 0 0 \n         0         0 1\n         1         1 2\n         2         2 3\n         3         3 4\n\nB=2 3\t\n\t   1   2   3\n\t   4   5   6\n\t\n'
        
def test_hello_world_verbosities(): 
    from pyfreefem.examples.ex12_verbosity_output import runner   
    runner.execute({'message':'Hello world',\
                    'SOLVE_LAPLACE':1},verbosity=1,\
                   plot=False)
   
    check = """1 : \n    2 : mesh Th=square(30,30);\n    3 : fespace Fh(Th,P1);\n    4 : Fh u,v;\n    5 : \n    6 : cout << "The value of SOLVE_LAPLACE is 1." << endl;\n    7 : solve laplace(u,v)=\n    8 :     int2d(Th)(dx(u)*dx(v)+dy(u)*dy(v))\n    9 :         -int2d(Th)(v)\n   10 :         +on(1,2,3,4,u=0);\n   11 : plot(u,cmm="Hello world");\n"""
    assert check in runner.rets[1]

    runner.execute({'message':'Hello world',\
                    'SOLVE_LAPLACE':1},verbosity=0,\
                   plot=False)
    assert runner.rets[1] == 'The value of SOLVE_LAPLACE is 1.\n'
        
test_ex11_io_import()



import numpy as np
from pyfreefem import FreeFemRunner, get_root, tic, toc
from pyfreefem.io import display
    
tic()   
np.random.seed(42)
big_matrix = np.random.rand(31758,397)
display("Generate matrix took "+toc()) 
    
code = r""" 
IMPORT "io.edp"  
        
tic; 
real[int,int] m = import2DArray("big_matrix"); 
disp("Import took "+toc); 
    
tic; 
export2DArray(m); 
disp("Export took "+toc); 
"""
runner = FreeFemRunner(code,run_dir="run",debug=2) 
tic()
exports = runner.import_variables(big_matrix)
display("Import took "+toc()) 
#
#    
tic()
exports = runner.execute(verbosity=0)
display("Execute took "+toc()) 
    
print("Big matrix = "+str(big_matrix))
print("FreeFEM = "+str(exports['m']))



# This file is part of PyFreeFEM.
#
# PyFreeFEM is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# PyFreeFEM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# A copy of the GNU General Public License is included below.
# For further information, see <http://www.gnu.org/licenses/>.
from pyfreefem.io import readFFMatrix, writeFFMatrix
from pyfreefem import FreeFemRunner
import numpy as np
import scipy as sp

code = """
mesh Th=square(1,1);
fespace Fh(Th,P1);
varf laplace(u,v)=int2d(Th)(2*u*v+dx(u)*dx(v)+dy(u)*dy(v));
matrix A = laplace(Fh,Fh,tgv=-2);
{
    ofstream f("$RUNDIR/ex8_A.gp");
    f << A;
}
cout << "FreeFEM : saved matrix ex8_A.gp" << endl;
cout << A << endl;
"""
with FreeFemRunner(code,debug=10) as runner: 
    runner.execute(verbosity=0)    
    A = readFFMatrix(runner.run_dir+"/ex8_A.gp")
print("Python : read matrix ex8_A.gp")
print(A)

# Sparse identity matrix
I = sp.sparse.eye(5,5)
code = """
matrix I;
{
    ifstream f("$RUNDIR/ex8_I.gp");
    f >> I;
}
cout << "FreeFEM : read matrix ex8_I.gp" << endl;
cout << I << endl;
"""
with FreeFemRunner(code) as runner: 
    writeFFMatrix(I, runner.run_dir+"/ex8_I.gp")
    print(I,"\n")
    print(f"\nPython : saved matrix {runner.run_dir}/ex8_I.gp")
    runner.execute()
    rets = runner.rets



# This file is part of PyFreeFEM.
#
# PyFreeFEM is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# PyFreeFEM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# A copy of the GNU General Public License is included below.
# For further information, see <http://www.gnu.org/licenses/>.
import re

from .io import colored, display
import os
    
"""PyFreeFEM parsing module"""

#: Enable or disable the parsing of double backslash ``\\``.   
ENABLE_LINE_BREAKS = True

MAGIC_CHARACTER = '$'
""" Prefix character for magic variables (by default ``$``)."""


class Preprocessor():
    """A class that parses FreeFEM ``.edp`` scripts with enhanced
    PyFreeFEM meta-language instructions (see :ref:`language`).
        
    The :class:`pyfreefem.preprocessor.Preprocessor` class is used by
    :class:`pyfreefem.FreeFemRunner` for parsing FreeFEM ``.edp`` scripts with enhanced
    preprocessing instructions and converting it into a proper, executable
    ``.edp`` code. The usage is similar to ``pyfreefem.FreeFemRunner``:
    
    .. code:: python
    
       from pyfreefem import Preprocessor
       # Parse a single file
       preproc=Preprocessor('solveState.edp')
    
       # Parse a list of files (to be assembled consecutively)
       preproc=Preprocessor(['params.edp','solveState.edp'])
    
       # Parse a single file with an updated configuration
        preproc=Preprocessor('solveState.edp',{'ITER':'0010'})
    
        # Parse raw code
       code = \"""DEFAULT (n,30)
       mesh Th=square($n,$n);
       preproc=Preprocessor(code)\"""
    
       # Then parse these files with `Preprocessor.parse':
       parsedCode=preproc.parse();
       parsedCode=preproc.parse({'n':'0001'});# Parse with a different configuration
    
    A configuration dictionary ``config`` containing the updated (or
    predefined) values of magic variables is assembled while parsing the
    ``.edp`` file.
        

    :param script:  single file name, list of
                    file names or raw ``.edp`` script

    :param config: (optional) a dictionary of magic variables with default values

    :param included: (optional) a list of file names that must not be
                included. Usually empty, used for implementation
                purposes.

    :param debug:  A tuning parameter for flexible verbose output, allowing to
                   check the behavior of the parsing.
    """
    ACTIONS = {'IF': '__execute_if',
               'ENDIF': '__execute_endif',
               'ELSE': '__execute_else',
               'IFEQ': '__execute_ifeq',
               'IFDEF': '__execute_ifdef',
               'FOR': '__execute_for',
               'ENDFOR': '__execute_endfor',
               'DEFAULT': '__execute_default',
               'SET':  '__execute_set',
               'DEBUG': '__execute_debug',
               'INCLUDE': '__execute_include',
               'SET_TEXTVAR': '__execute_set_textvar',
               'END_TEXTVAR': '__execute_end_textvar',
               'IMPORT':'__execute_import'}

    def __init__(self, script, config=dict(), included=set(), debug=0):
        self.code = []
        self.files = []
        self.debug = debug
        self.interpreted = ''
        self.state = []
        self.position = 0

        #: Dictionary of magic variables.   
        #: This dictionary is updated after calling :meth:`pyfreefem.preprocessor.Preprocessor.parse`.
        self.config = dict()
        self.stackFor = dict()
        self.imports = []
        if isinstance(script, str):
            self.files = [script]
        else:
            self.files = script
        self.initConfig = config
        for fileName in self.files:
            if len(fileName.splitlines())==1:
                display("Including "+fileName, 1, debug, 'green')
                try:
                    with open(fileName, "r") as f:
                        self.code = self.code+f.readlines()
                except:
                    display("Warning: it seems "+fileName+" is a file name "\
                            "however this file does not exist. Will treat as "\
                            "portion of code.",0,self.debug, "red")
                    self.code = self.code+[x+'\n' for x in fileName.splitlines()]
                    self.code[-1] = self.code[-1][:-1]  # Remove extra \n
            else:
                self.code = self.code+[x+'\n' for x in fileName.splitlines()]
                self.code[-1] = self.code[-1][:-1]  # Remove extra \n
        self.initialIncluded = set(included)

    def replace(self, string, config=None):
        """Replace a string containing magic variables with their values provided   
           by the ``config`` parameter.

           :param string: string to replace
           :param config: (default: None). The magic variables are replaced      
                          according to the values given by `config`
                          If not specified, config is set to be the attribute
                          :attr:`pyfreefem.preprocessor.Preprocessor.config`.
        """
        if config is None:
            config = self.config
        keys = list(config.keys())
        keys.sort(reverse=True)
        ret = string
        for key in keys:
            replacement = str(config[key])
            if os.name == 'nt':
                replacement = replacement.replace('\\','/')
            ret = ret.replace(MAGIC_CHARACTER+'{'+key+'}', replacement)
            ret = ret.replace(MAGIC_CHARACTER+key, replacement)
        return ret

    def parse(self, config=dict()):
        """Parse the ``.edp`` script provided to the :class:`pyfreefem.preprocessor.Preprocessor`   
        instance.
            
        :param config: a configuration for the magic variable that takes precedence over    
                       the default ``config`` value provided to the 
                       :class:`pyfreefem.preprocessor.Preprocessor` constructor. 

                       .. code:: python

                          preproc = Preprocessor('file.edp')
                          parsedCode=preproc.parse({'hmin':'0.002'})

        Once the :meth:`pyfreefem.preprocessor.Preprocessor.parse` has been called,  
        the parameter :attr:`pyfreefem.preprocessor.Preprocessor.config` is   
        updated.    
        """
        # Set the cursor at position 0 and initialize the configuration
        # with the values supplied at initialization
        self.position = 0
        self.interpreted = ''
        self.state = []
        self.imports = []
        self.stackFor = dict()
        self.config = dict(self.initConfig)
        self.config.update(config)
        self.included = set(self.initialIncluded)
        # Read each line of the file and interpret.
        # For each line, either we have an instruction, or some code that is
        #interpreted and store in self.interpreted
        while self.position < len(self.code):
            line = self.replace(self.code[self.position], config=self.stackFor)
            display(str(self.position)+" : "+line.strip(), 2, self.debug)
            instruction = line.split()
            if instruction != [] and instruction[0] in self.ACTIONS:
                exec(f"self._Preprocessor{self.ACTIONS[instruction[0]]}"
                     "(instruction)")
            else:
                self.__read_line()
                self.position = self.position+1
        if self.state != []:
            raise Exception("Parse error : the state stack is not empty.\n"
                            + self.state.__str__())
        # Remove magic comments
        self.interpreted = re.sub(r'(\n\s*\/\/\*\*.*)+\n',
                                  r'\n', self.interpreted)
        self.interpreted = re.sub(r'\/\/\*\*.*', '', self.interpreted)
        # Arrange line breaks with \\ characters
        if ENABLE_LINE_BREAKS:
            self.interpreted = re.sub(r'\\\\\n\s*', '', self.interpreted)
            self.interpreted = re.sub(r'\n\s*\\\\', '', self.interpreted)
        # Remove extra lines
        self.interpreted = re.sub(r'\n(\n)+\n', r'\n\n', self.interpreted)
        return self.interpreted

    def __read_line(self):
        """Read a line which is not a special instruction."""
        # Starts by checking if line must be added depending on whether
        # we are within a if or for loop, or text variable
        ignore = False
        try:
            if self.state[-1][1] == 'if' \
                    and self.state[-1][2] in {False, 'ignore'}:
                ignore = True
        except:
            pass
        try:
            if self.state[-1][1] == 'for' \
                    and self.state[-1][3] >= self.state[-1][4]:
                ignore = True
        except:
            pass
        try:
            if self.state[-1][1] == 'for' and self.state[-1][2] == 'ignore':
                ignore = True
        except:
            pass
        line = self.code[self.position]
        line = self.replace(line, config=self.stackFor)
        line = self.replace(line)
        try:
            if self.state[-1][1] == 'set_text_var':
                self.state[-1][3] += line
                ignore = True
        except:
            pass
        if not ignore:
            self.interpreted = self.interpreted+line
            if self.debug > 1:
                display(colored("Interpreted : ", "green") +
                      colored(line[:-1], "blue"),2,self.debug)

    def __execute_include(self, instruction):
        """Include file"""
        fileToInclude = self.replace(instruction[1][1:-1])
        if not fileToInclude in self.included:
            self.included.add(fileToInclude)
            t = Preprocessor(fileToInclude, self.config, self.included)
            t.debug = self.debug
            self.interpreted = self.interpreted+t.parse()
            self.config.update(t.config)
            # Avoid inclusion of already included files
            self.included = self.included.union(t.included)
        self.position = self.position+1

    def __execute_if(self, instruction):
        var = instruction[1]
        ignore = False
        try:
            if self.state[-1][1] == 'if' and not self.state[-1][2]:
                ignore = True
            if self.state[-1][2] == 'ignore':
                ignore = True
        except:
            pass
        if ignore:
            result = 'ignore'
        else:
            if var in self.config:
                result = True
                try:
                    if str(self.config[var]) == '0':
                        result = False
                except:
                    pass
            else:
                result = False
        self.state.append([self.position, 'if', result])
        self.position = self.position+1
        display("State : "+self.state[-1].__str__(), 1, self.debug, 'green')

    def __execute_endif(self, instruction):
        if self.state[-1][1] != 'if':
            raise Exception("Parse error at line "+str(self.position))
        display("Closed : "+self.state[-1].__str__(), 1, self.debug, 'green')
        self.state.pop()
        self.position = self.position+1

    def __execute_else(self, instruction):
        ignore = False
        try:
            if self.state[-1][1] == 'if' and self.state[-1][2] == 'ignore':
                ignore = True
            if self.state[-1][1] == 'for' and self.state[-1][2] == 'ignore':
                ignore = True
        except:
            pass
        if not ignore:
            self.state[-1][2] = (not self.state[-1][2])
        self.position = self.position+1
        display("State : "+self.state[-1].__str__(), 1, self.debug, 'green')

    def __execute_ifeq(self, instruction):
        instruction = " ".join(instruction)
        instruction = re.findall(
            r'\[[^\]]*\]|\([^\)]*\)|\"[^\"]*\"|\S+', instruction)
        for i in range(1, len(instruction), 2):
            instr = instruction[i]
            m = re.search(r'\(\s*(.+)\s*,\s*"(.+)"\s*\)', instr)
            if m:
                var = m.group(1)
                toCompare = m.group(2)
                result = var in self.config and str(
                    self.config[var]) == toCompare
                # Allow to test for variables
                result = result or var in self.stackFor and str(
                    self.stackFor[var]) == toCompare
            else:
                m = re.search(r'\(\s*(.+)\s*,\s*\[(.+)\]\s*\)', instr)
                var = m.group(1)
                toCompare = set([s[1:-1] for s in m.group(2).split(',')])
                result = var in self.config and str(
                    self.config[var]) in toCompare
            if result:
                break
        for i in range(2, len(instruction), 2):
            if not instruction[i] == 'OR':
                raise Exception(
                    "Parse error in IFEQ, for now only OR "
                    "is supported between statements.")
        try:
            if self.state[-1][1] == 'if' \
                    and self.state[-1][2] in {False, 'ignore'}:
                result = 'ignore'
            if self.state[-1][1] == 'for' \
                    and self.state[-1][2] == 'ignore':
                result = 'ignore'
        except:
            pass
        self.state.append([self.position, 'if', result])
        self.position = self.position+1
        display("State : "+self.state[-1].__str__(), 1, self.debug, 'green')

    def __execute_ifdef(self, instruction):
        var = instruction[1]
        result = var in self.config
        try:
            if self.state[-1][1] == 'if' \
                    and self.state[-1][2] in {False, 'ignore'}:
                result = 'ignore'
            if self.state[-1][1] == 'for' \
                    and self.state[-1][2] == 'ignore':
                result = 'ignore'
        except:
            pass
        self.state.append([self.position, 'if', result])
        self.position = self.position+1
        display("State : "+self.state[-1].__str__(), 1, self.debug, 'green')

    def __execute_for(self, instruction):
        ignoreFor = False
        try:
            if self.state[-1][1] == 'if' \
                    and self.state[-1][2] in {False, 'ignore'}:
                self.state.append([self.position, 'for', 'ignore'])
                ignoreFor = True
        except:
            pass
        if not ignoreFor:
            instr = instruction[1]
            m = re.search(r'\((.+),(.+),(.+),"(.+)"\)', instr)
            if not hasattr(m, 'group'):
                m = re.search(r'\((.+),(.+),(.+)\)', instr)
            try:
                start = int(m.group(2))
            except ValueError:
                start = int(self.config[m.group(2)])
            try:
                end = int(m.group(3))
            except ValueError:
                if m.group(3) in self.config:
                    end = int(self.config[m.group(3)])
                else:
                    end = 0
            if start >= end:
                self.state.append([self.position, 'for', 'ignore'])
            else:
                self.state.append(
                    [self.position, 'for', m.group(1), start, end])
                if m.lastindex == 4:
                    self.state[-1].append(m.group(4))
                else:
                    self.state[-1].append('')
            self.stackFor[m.group(1)] = start
            display("Assigned for variable : " +
                    m.group(1)+"="+str(start), 1, self.debug, 'green')
        self.position = self.position+1
        display("State : "+self.state[-1].__str__(), 1, self.debug, 'green')

    def __execute_endfor(self, instruction):
        if self.state[-1][1] != 'for':
            raise Exception("Parse error at line "+self.position)
        if self.state[-1][2] == 'ignore':
            self.position = self.position+1
            self.state.pop()
        else:
            varFor = self.state[-1][2]
            if self.stackFor[varFor] >= self.state[-1][4]-1:
                self.stackFor.pop(varFor)
                self.state.pop()
                self.position = self.position+1
            else:
                self.interpreted = self.interpreted+self.state[-1][5]
                self.stackFor[varFor] = self.stackFor[varFor]+1
                display("Assigned for variable : "+varFor +
                        "="+str(self.stackFor[varFor]), 1, self.debug, 'green')
                self.position = self.state[-1][0]+1
        if self.debug > 0:
            try:
                display("State : "+self.state[-1].__str__(), 1, self.debug, 'green')
            except:
                pass

    def __execute_default(self, instruction):
        try:
            if self.state[-1][1] == 'if' and self.state[-1][2] in {False,
                                                                   'ignore'}:
                self.position = self.position+1
                return
        except:
            pass
        instruction[1] = " ".join(instruction[1:])
        m = re.search(r'\((.+)\s*,\s*"(.+)"\)', instruction[1])
        try:
            key = m.group(1)
            value = m.group(2)
            if not key in self.config:
                value = self.replace(value)
                self.config[key] = value
        except:
            try:
                m = re.search(r'\((.+)\s*,\s*(.+)\)', instruction[1])
                key = m.group(1)
            except:
                raise Error("Error while parsing DEFAULT instruction")
            value = m.group(2)
            if not key in self.config:
                value = eval(self.replace(value))
                self.config[key] = value
        display("Defined "+key+"="+str(self.config[key]), 1, self.debug, 'green')
        self.position = self.position+1

    def __execute_set(self, instruction):
        try:
            if self.state[-1][1] == 'if' and self.state[-1][2] in {False,
                                                                   'ignore'}:
                self.position = self.position+1
                return
        except:
            pass
        instruction[1] = " ".join(instruction[1:])
        m = re.search(r'\(\s*(.+)\s*,\s*"(.+)"\s*\)', instruction[1])
        try:
            key = m.group(1)
            value = m.group(2)
            value = self.replace(value)
        except:
            try:
                m = re.search(r'\((.+)\s*,\s*(.+)\)', instruction[1])
                key = m.group(1)
            except:
                raise NameError("Error while parsing SET instruction")
            value = m.group(2)
            value = eval(self.replace(value))
        self.config[key] = value
        display("Defined "+key+"="+str(self.config[key]), 1, self.debug, 'green')
        self.position = self.position+1

    def __execute_set_textvar(self, instruction):
        try:
            if self.state[-1][1] == 'if' \
                    and self.state[-1][2] in {False, 'ignore'}:
                self.position = self.position+1
                return
        except:
            pass
        nameVar = instruction[1]
        self.state.append([self.position, 'set_text_var', nameVar, ''])
        self.position = self.position+1
        display("State : "+self.state[-1].__str__(), 1, self.debug, 'green')

    def __execute_end_textvar(self, instruction):
        try:
            if self.state[-1][1] == 'if' \
                    and self.state[-1][2] in {False, 'ignore'}:
                self.position = self.position+1
                return
        except:
            pass
        if self.state[-1][1] != 'set_text_var':
            raise Exception("Parse error at line "+str(self.position))
        display("Closed : "+self.state[-1].__str__(), 1, self.debug, 'green')
        state = self.state.pop()
        self.set({state[2]: state[3]})
        display("Set : "+state[2]+":\n"+state[3], 1, self.debug, 'green')
        self.position = self.position+1

    def __execute_debug(self, instruction):
        import ipdb;
        ipdb.set_trace();
        self.position = self.position + 1

    def __execute_import(self, instruction):
        macro_file = os.path.dirname(__file__)+"/edp/"+instruction[1][1:-1]
        self.imports.append(macro_file)
        if not os.path.exists(macro_file):
            raise NameError("Error while parsing IMPORT instruction: file \""+macro_file+"\" does not "
                        "exist")
        
        filename = instruction[1][1:-1]
        line = "include \""+self.config['RUNDIR']+"/"+filename+"\"\n"
        self.interpreted += line
        if self.debug > 1:
            display(colored("Interpreted : ", "green") +
                  colored(line[:-1], "blue"),2,self.debug)
        self.position += 1

        
    def set(self, assign):
        """Update the config dictionary 
            
        :param assign: New values for the dictionary :attr:`pyfreefem.preprocessor.Preprocessor.config`.
        :type assign: dict"""
        self.config.update(assign)



# This file is part of PyFreeFEM.
#
# PyFreeFEM is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# PyFreeFEM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# A copy of the GNU General Public License is included below.
# For further information, see <http://www.gnu.org/licenses/>.

from .freefemrunner import FreeFemRunner
from .preprocessor import Preprocessor
from .io import readFFArray, readFFMatrix, writeFFArray, writeFFMatrix, extract
from .io import make_dictionary, get_root, tic, toc



# This file is part of PyFreeFEM.
#
# PyFreeFEM is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# PyFreeFEM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# A copy of the GNU General Public License is included below.
# For further information, see <http://www.gnu.org/licenses/>.
from pyfreefem import FreeFemRunner
try:
    from pymedit import Mesh, P1Function
    WITH_PYMEDIT = True
    import matplotlib.pyplot as plt
    plt.ion()
except:
    WITH_PYMEDIT = False


import os
examples_folder = os.path.split(__file__)[0]

with FreeFemRunner(examples_folder+"/edp/ex9.edp") as runner:
    # FreeFemRunner will save files in a temporary directory
    # accessible in runner.script_folder (python) or $RUNDIR (in FreeFEM)
    # runner and The temporary directory is deleted outside the with context
    runner.execute({'n':40})
    if WITH_PYMEDIT:
        M = Mesh(runner.run_dir+"/Th.mesh")
        phi = P1Function(M,runner.run_dir+"/phi.sol")
    with open(runner.run_dir+"/result.gp","r") as f:
        result=float(f.readlines()[0].strip())


if __name__=="__main__":
    print("Result=",result)
    if WITH_PYMEDIT:
        M.plot(title="Mesh Th")
        phi.plot(title="P1Function f")
        input("Press any key to close the plots.")

    import matplotlib.pyplot as plt
    plt.close('all')



# Copyright 2018-2019 CNRS, Ecole Polytechnique and Safran.
#
# This file is part of pyfreefem.
#
# nullspace_optimizer is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# nullspace_optimizer is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# A copy of the GNU General Public License is included below.
# For further information, see <http://www.gnu.org/licenses/>.

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyfreefem",
    version="1.1.4",
    author="Florian Feppon",
    author_email="florian.feppon@kuleuven.be",
    license="GNU GPL version 3",
    description="Package PyFreeFEM for interfacing Python and FreeFEM.",
    keywords="FreeFEM",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://pyfreefem.readthedocs.io/en/latest/",
    packages=setuptools.find_packages(),
    package_data={'':['*.edp','*/*.edp']},    
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Mathematics"
    ],
    install_requires=['numpy>=1.12.1',
                      'scipy>=0.19.1',  
                      'colored>=1.3.93',    
                      'pymedit>=1.2'],
    python_requires='>=3.6',
)



# This file is part of PyFreeFEM.
#
# PyFreeFEM is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# PyFreeFEM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# A copy of the GNU General Public License is included below.
# For further information, see <http://www.gnu.org/licenses/>.
from pyfreefem import Preprocessor

import os

examples_folder = os.path.split(__file__)[0]
preproc = Preprocessor(examples_folder+"/edp/ex4.edp")
    
parsed = preproc.parse()
print(parsed)
